import cv2
from pyzbar.pyzbar import decode


def read_qr(imgpath):
    im = cv2.imread(imgpath)

    codes = decode(im)  # auto detect code type

    all_codes = []

    if codes:
        for code in codes:
            all_codes.append(code.data.decode("utf-8"))

    return all_codes



import base58


_qr_charset = b"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:"
_confiks_charset = b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHI"


def raw_decoder(raw):
    """
    The QR code uses a custom base45 decoder as written by confiks on GitHub:https://github.com/confiks/base45-go/
    This decoder manages to cram a few extra bytes into the QR code compared to a more traditional base45 encoder.

    The confiks decoder works similar to a base58 decoder but with a reduced alphabet.


    :return: ASN.1 blob
    """

    return base58.b58decode(_qr_to_confiks(raw), alphabet=_confiks_charset)



def _qr_to_confiks(data):
    """
    This converts the QR code to confiks alphabet.
    """
    qr_charset_len = 45

    qr_confiks = {}

    for i in range(0, qr_charset_len):
        qr_confiks[_qr_charset[i]] = _confiks_charset[i]

    input_len = len(data)
    confiks_encoded_input = []

    for i in range(0, input_len):
        confiks_encoded_input.append(qr_confiks[data[i]])

    return bytes(confiks_encoded_input)



from pathlib import Path
from cleo import Command, Application

from coronacheck_tools import decode_qr
from coronacheck_tools.certificate_versions.v2 import v2_asn1_specs

import pkg_resources  # part of setuptools
import json


AFFILIATE_WARNING = 'This is an unofficial tool that is in no way affiliated with CoronaCheck.nl or the Ministry of VWS'


class Asn1SpecCommand(Command):
    """
    Show ASN.1 spec for a specified version.

    asn1spec
        {version=2 : Currently supported versions: 2}
    """

    def handle(self):
        version = int(self.argument('version'))

        if version == 2:
            print(v2_asn1_specs)


class DumpCommand(Command):
    """
    Dump the QR Code data from image to RAW, ASN.1 or JSON.

    dump
        {format : Output format. RAW, ASN1, JSON.}
        {image : Path to an image file with one or more QR codes.}
        {output : Output directory. Will overwrite existing files.}
    """

    def handle(self):
        format = self.argument('format').upper().strip()
        image_path = Path(self.argument('image'))
        output_path = Path(self.argument('output'))

        valid_formats = ('RAW', 'ASN1', 'JSON')
        error = False
        if format not in valid_formats:
            self.line(f'<error>Invalid output format: {format} specify one of {", ".join(valid_formats)}</error>')
            error = True

        if not image_path.is_file():
            self.line(f'<error>Input image file does not exist: {image_path}</error>')
            error = True

        if not output_path.is_dir():
            self.line(f'<error>Output directory does not exist: {output_path}</error>')
            error = True

        if error:
            return

        if format == 'ASN1':
            format = 'ASN1_BLOB'  # retrieve binary ASN1 blob

        if format == 'JSON':
            format = 'DICT'

        retvals = decode_qr(image_path, format=format)

        mode = 'w'
        if format == 'RAW':
            extension = '.raw'
        elif format == 'ASN1_BLOB':
            extension = '.asn'
            mode = 'wb'
        elif format == 'DICT':
            retvals = [json.dumps(x) for x in retvals]
            extension = '.json'

        counter = 0
        for data in retvals:
            fname = output_path / f'dhc_data_{counter:03}{extension}'
            print(f'Writing code {counter+1}/{len(retvals)} to {fname}')
            with open(fname, mode) as fh:
                fh.write(data)
            counter += 1

def main():
    application = Application(name="coronacheck-tools", version=pkg_resources.require("coronacheck-tools")[0].version)
    application.add(DumpCommand())
    application.add(Asn1SpecCommand())

    print(AFFILIATE_WARNING)

    application.run()



import asn1tools


v2_asn1_specs = """
DHC DEFINITIONS ::= BEGIN
    ProofSerializationV2 ::= SEQUENCE {
        disclosureTimeSeconds  INTEGER,
        c                      INTEGER,
        a                      INTEGER,
        eResponse              INTEGER,
        vResponse              INTEGER,
        aResponse              INTEGER,
        aDisclosed             SEQUENCE OF INTEGER
    }
    
    CredentialMetadataSerialization ::= SEQUENCE {
        -- CredentialVersion identifies the credential version, and is always a single byte
        credentialVersion OCTET STRING,

        -- IssuerPkId identifies the public key to use for verification
        issuerPkId PrintableString
    }
END
"""

v2_asn1_decoder = asn1tools.compile_string(v2_asn1_specs)


def v2_dhc_records_to_dict_repr(dhc_records):
    retval = dhc_records.copy()


    retval['aDisclosed'][0] = _decode_metadata_attrs(dhc_records['aDisclosed'][0])

    for i in range(1, len(retval['aDisclosed'])):
        retval['aDisclosed'][i] = _int_to_byte_attrs(dhc_records['aDisclosed'][i]).decode('utf-8')

    retval['aDisclosed'] = _attrs_to_dict(retval['aDisclosed'])
    retval['aDisclosed']['CredentialMetadata']['credentialVersion'] = retval['aDisclosed']['CredentialMetadata']['credentialVersion'].hex()


    return retval


def _attrs_to_dict(lst):
    return {
        "CredentialMetadata": lst[0],
        "isSpecimen": lst[1],
        "isPaperProof": lst[2],
        "validFrom": lst[3],
        "validForHours": lst[4],
        "firstNameInitial": lst[5],
        "lastNameInitial": lst[6],
        "birthDay": lst[7],
        "birthMonth": lst[8],
    }


def _int_to_byte_attrs(int_input):
    if int_input & 1 == 0:
        return []
    else:
        h = hex(int_input >> 1).split('x')[1]
        if len(h) == 1:
            h = f'0{h}'

        # print(h)

        return bytes.fromhex(h)


def _decode_metadata_attrs(record):
    attribute_bytes = _int_to_byte_attrs(record)

    credentials_metadata = v2_asn1_decoder.decode('CredentialMetadataSerialization', attribute_bytes)

    # if credentials_metadata['credentialVersion'] == b'\x02':

    return credentials_metadata



from pathlib import Path
from coronacheck_tools.common.qrreader import read_qr
from coronacheck_tools.common.decoders import raw_decoder
from coronacheck_tools.certificate_versions.v2 import v2_asn1_decoder, v2_dhc_records_to_dict_repr

import asn1tools


supported_versions = (2,)
v2_valid_output_formats = ['RAW', 'ASN1_BLOB', 'ASN1', 'DICT']


def _check_step(format, step, version):
    if version not in supported_versions:
        raise ValueError(f'Version {version} unsupported. Only version(s) {", ".join(supported_versions)} are currently supported.')

    if version == 2:
        valid_output_formats = v2_valid_output_formats

    format = format.upper()
    if format not in valid_output_formats or valid_output_formats.index(format) < valid_output_formats.index(step):
        raise ValueError(f'Invalid format, choose one of {", ".join(valid_output_formats[valid_output_formats.index(format) + 1:])}')

    return format


def decode_qr(image_file, format='dict'):
    format = _check_step(format, 'RAW', 2)  # default to v2, it doesn't really matter for decoding the QR
    image_file = Path(image_file)

    if not image_file.is_file():
        raise ValueError(f'image_file {image_file} does not exist')

    qr_codes = read_qr(str(image_file))

    if format == 'RAW':
        return qr_codes

    qr_codes = [qr for qr in qr_codes if len(qr) > 4 and qr.startswith('NL') and qr[3] == ':']

    return [decode_raw(qr, format=format, version=int(qr[2], 16)) for qr in qr_codes]


def decode_raw(raw, format='dict', version='auto'):
    if not (raw.startswith('NL') and raw[3] == ':'):
        raise ValueError(f'Invalid data. RAW data should start with NL: {raw}')

    if version == 'auto':
        version = int(raw[2], 16)

    format = _check_step(format, 'ASN1_BLOB', version)

    if version == 2:
        data = raw[4:]

    asn1_blob = raw_decoder(data.encode())

    if format == 'ASN1_BLOB':
        return asn1_blob

    return decode_asn1_blob(asn1_blob, format=format, version=version)


def decode_asn1_blob(asn1_blob, format='dict', version=2):
    format = _check_step(format, 'ASN1', version)

    if version == 2:
        dhc_records = v2_asn1_decoder.decode('ProofSerializationV2', asn1_blob)

    if format == 'ASN1':
        return dhc_records

    return decode_to_dict(dhc_records)


def decode_to_dict(dhc_records, format='dict', version=2):
    format = _check_step(format, 'DICT', version)

    if version == 2:
        return v2_dhc_records_to_dict_repr(dhc_records)



from setuptools import setup, find_packages
import sys

if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported')

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='coronacheck-tools',
    version='1.0.0',
    description='Unofficial tool to dump QR code data from CoronaCheck.nl to RAW, ASN.1 or JSON. Not affiliated with CoronaCheck.nl or Ministry of VWS.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Thomas Phil',
    author_email='thomas@tphil.nl',
    url='https://github.com/Sikerdebaard/coronacheck-tools',
    python_requires=">=3.6",
    packages=find_packages(),  # same as name
    install_requires=[
        'cleo>=0.8.1',
        'base58>=2.1.0',
        'pyzbar>=0.1.8',
        'opencv-python>=4.5.2.54',
        'asn1tools>=0.158.0',
    ],
    entry_points={
        'console_scripts': [
            'coronacheck-tools=coronacheck_tools.cli:main',
        ],
    },
)
